import java.io.*;
import java.util.*;

public class DPDecodeRunner {

	public static String expandUTF8(String str) {
		// expand unicode characters to utf-8
		// multi-character representations

		// to solve the issue that a sentence received by
		// Java as a command-line argument has non-ASCII
		// utf-8 characters converted to single Unicode
		// characters in the string, whereas all the rest
		// of the code works with the multi-byte
		// representations of such characters

		try {
			byte [] utf8bytes = str.getBytes("UTF-8");
			String utf8str = new String(utf8bytes);
			//System.out.println("utf-8-ified: " + utf8str);
			return utf8str;
		} catch (UnsupportedEncodingException e) {
			System.out.println("UTF-8 CONVERSION FAILED!");
			return "";
		}
	}

    public static void main(String [] args) {
	if (args.length != 3) {
	    System.err.println("Usage: DPDecodeRunner [1|2|3] inputSentence numResults");
	    return;
	}

	String model = args[0];
	String input = expandUTF8(args[1]);
	// String input = "también recibió dicha información";
	int numResults = 1;
	try {
	    numResults = Integer.parseInt(args[2]);
	} catch (NumberFormatException e) {
	    System.err.println("Invalid number of results: "+args[2]);
	}

	// TODO: a class to convert from a lang pair to the following files

	// Spanish-English:
	String sourceTextFile = "../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.es";
        String destTextFile = "../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.en";
        String [] paramsTextFiles = {"../parameters/textfiles/europarl_2-to-10_stripped.training.es-en.model1.5iters",
                                     "../parameters/textfiles/europarl_2-to-10_stripped.training.es-en.model2.5iters",
                                     "../parameters/textfiles/model3-iter-patched-smoothing-3"};
        String sbpFile = "../language_models/europarl_2-to-10_stripped.training.es-en.en.ntp.sbp";
        String srilmFile = "../language_models/europarl_2-to-10_stripped.training.es-en.en.lm";

	// French-English:
	//String sourceTextFile = "../corpora/textpairs/europarl_2-to-10_stripped.training.fr-en.fr";
        //String destTextFile = "../corpora/textpairs/europarl_2-to-10_stripped.training.fr-en.en";
        //String paramsTextFile = "../parameters/textfiles/europarl_2-to-10_stripped.training.fr-en.model2.5iters";
        //String sbpFile = "../language_models/europarl_2-to-10_stripped.training.fr-en.en.ntp.sbp";
        //String srilmFile = "../language_models/europarl_2-to-10_stripped.training.fr-en.en.lm";

	ParametersFromTextFile params;// = new ParametersFromTextFile(paramsTextFile);
        CorpusFromTextPair corpus = new CorpusFromTextPair(sourceTextFile, destTextFile);
        BigramsFromSrilmFile bfsf = new BigramsFromSrilmFile(srilmFile);
        SortedProbabilityAccessor nextToPrevs = new SortedProbabilityAccessor(sbpFile);
	DPDecoder dec;
	if (model.equals("1")) {
	    params = new ParametersFromTextFile(paramsTextFiles[0]);
	    dec = new DPModel1Decoder(params, corpus, bfsf, nextToPrevs);
	} else if (model.equals("2")) {
	    params = new ParametersFromTextFile(paramsTextFiles[1]);
	    dec = new DPModel2Decoder(params, corpus, bfsf, nextToPrevs);
	} else if (model.equals("3")) {
	    params = new ParametersFromTextFile(paramsTextFiles[2]);
	    dec = new DPModel3Decoder(params, corpus, bfsf, nextToPrevs);
	} else {
	    System.err.println("Invalid model: " + model);
	    return;
	}

	// System.out.println("==== Testing Unknown Word Detection ====");
	// String test = "U0 consejeros U1 de seguridad U2 U3 para U4 U5 U6 el transporte de U2 U3 mercancías U1 U2 peligrosas U9";
	
	// dec.detectUnknownWords(test);
	// System.out.println("==== Finished ====");

	//System.out.println("Decoding: " + input + " ...");

	String [] results = dec.decode(input, numResults);
	if(results.length > 0) {
		for (int i = 0; i < results.length; i++) {
	    	System.out.println(results[i]);
		}
	} else {
		System.out.println("NO RESULTS FOUND");
	}

	// temporary for evaluating results of speedup tricks
	//ArrayList<Pair<String, Double>> results = dec.decodeToLength("esto ya lo hemos escuchado antes", 7, 100);
    }
}
