public class DPModel3Decoder extends DPDecoder {

    protected double [] factorials;

    public static final int maxFertility = 32;

    public DPModel3Decoder(ParameterAccessor p, CorpusAccessor c, LanguageModel lm, SortedProbabilityAccessor ntp) {
	super(p, c, lm, ntp);

	// Model 3 uses the P param to punish non-covered (spurious) words
	requireFullCoverage = false;

	factorials = new double[maxFertility];
	factorials[0] = 0.0;
	for(int i = 1; i < factorials.length; i++) {
		factorials[i] = factorials[i - 1] + (logTen(i));
		//System.out.println(Integer.toString(i) + "! = " + Double.toString(factorials[i]));
	}
    }

    protected double getKnownFertilityProb(int l, String e) {
    	double factorialPart = 0.0; //full log probability
    	if(l >= maxFertility) {
    		factorialPart = factorials[maxFertility - 1];
    	} else if(l > 0) {
    		factorialPart = factorials[l];
    	}

    	return params.getLogN(l, e) + factorialPart;
    }
    
    protected double getAlignmentProb(int i, int j, int I, int J) {
	return params.getLogD(j, i, I, J);
    }

    protected double getSentenceEndProb(String e, String [] sourceSentence, int coverageBitField) {
    	double model3Contribution = 0.0;

    	int m = sourceSentence.length; //length of the translation-source sentence
    	assert(m < maxFertility);
    	int phi0 = 0; //number of spurious words
    	for(int j = 1; j <= m; j++) {
    		if((coverageBitField & (1 << (j-1))) == 0) {
    			//this position is uncovered
    			phi0++;
    		}
    	}

    	if(m - phi0 < phi0) {
    		//then we have more than one non-spurious word per
    		// each spurious word in the source sentence;
    		// model 3 doesn't allow this
    		return Double.NEGATIVE_INFINITY;
    	}

    	double baseTenP = params.getP();

    	double p0 = logTen(1.0 - baseTenP); //prob. of not inserting spurious
    	double p1 = logTen(baseTenP); //prob. of inserting spurious

    	// account for binomial coefficient thing involving p0, p1, m, and phi0

    	//find m - phi0 choose phi0
    	double combination = factorials[m - phi0] - factorials[m - 2 * phi0] - factorials[phi0];

    	model3Contribution += combination + (m - 2 * phi0) * p0 + phi0 * p1;


    	//account for dividing by factorial of phi0

    	model3Contribution -= factorials[phi0];


    	//account for t(f_j | NULL)

    	for(int j = 1; j <= m; j++) {
    		if((coverageBitField & (1 << (j-1))) == 0) {
    			//this position is uncovered
    			model3Contribution += getTranslationProb(sourceSentence[j - 1], "_N_U_L_L_");
    		}
    	}

    	return super.getSentenceEndProb(e, sourceSentence, coverageBitField) + model3Contribution;
    }

    public static double logTen(double val) {
    	return Math.log(val) / Math.log(10);
    }
}
