import java.net.*;
import java.io.*;
import java.nio.ByteBuffer;
import java.util.Arrays;

class LMClient implements LanguageModel {
	Socket socket;
	InputStream in;
	PrintWriter output;
	int responseSize = 14;
	
	public static void main(String[] args) throws IOException {
		LMClient c = new LMClient();
		if (args.length == 3) {
		    c.open(args[1], Integer.parseInt(args[2]));
		}
		else {
		    c.open("127.0.0.1", 50000);
		}
		for (int i = 0; i < 1; i++) {
			System.out.println(c.queryNgram("Thank you"));

		}
		c.close();

	
	}
	
	public void open(String addr, int port) throws IOException {
		socket = new Socket(addr, port);
		in = socket.getInputStream();
		output = new PrintWriter(socket.getOutputStream(), true);
		
		byte[] serverMessages = new byte[17];
		int length = 17;
		int offset = 0;
		while (offset < length) {
			offset = in.read(serverMessages, offset, length);
		}
		
		output.println("_R_E_M_O_T_E_L_M_V=2");
	
		length = 4;
		offset = 0;
		while (offset < length) {
			offset = in.read(serverMessages, offset, length);
		}
	}
	
	public double queryNgram(String ngram) {
		String message = "W " + ngram;
		output.println(message);
		byte[] response = new byte[responseSize];
		int newOffset = 0;
		int offset = 0;
		int lastByte = 2;
		try {
    		in.skip(3);
    	} catch(IOException e) {
    	    System.err.println("IOException in LMClient queryNgram!!!");
    	    return 0.0;
    	}
		while (response[lastByte-2] != 10) {
			offset = lastByte-2;
			try {
			    newOffset = in.read(response, offset, responseSize);
			} catch(IOException e) {
			    System.err.println("IOException in LMClient queryNgram!!!!!");
			    return 0.0;
			}
			lastByte = offset+newOffset;	
			
			/*System.out.print(lastByte);
			System.out.print("\t");
			System.out.println(response[lastByte-2]);
			System.out.println(new String(response));*/
			//System.out.println(new String(response));
			//System.out.println(response[lastByteIndex-2]);
			//System.out.println(lastByteIndex-2);
		}		
		String strResponse = new String(response);
		//System.out.println(strResponse.length());
		if (strResponse.substring(0,4).equals("-inf")){
			//System.out.println(strResponse.length());
			return 0.0;
		}
		return Double.parseDouble(new String(response));
	}
	
	public double querySentence(String sentence) throws IOException {
        //TODO generalize for tri-grams, etc

		String[] words = sentence.split(" ");
		return querySentence(words);

	
	}
	
	public double querySentence(String[] words) throws IOException {
	    String[] sentence = new String[words.length+2];
        sentence[0] = "<s>";
        sentence[words.length+1] = "</s>";
		for (int i = 1; i < words.length+1; i++) {
		    sentence[i] = words[i-1];
		}
		
		int length = sentence.length;
		double sentenceProb = 0.0; 
		for (int i = 0; i < length - 1; i++) {
			String ngram = sentence[i] + " " + sentence[i+1];
			//System.out.println(ngram);
			double prob = this.queryNgram(ngram);
			//System.out.println(prob);
			sentenceProb += prob;
		}
		return sentenceProb;	
    }
	
	public void close() throws IOException {
		socket.close();
		output.close();
		in.close();
	}


}