import java.io.*;
import java.net.*;
import java.nio.ByteBuffer;

public class LanguageModelClient {
    
    private static final int TIMEOUT = 3000;
    
    private static final int MAXTRIES = 5;
    
    private int port;
    
    private InetAddress serverAddress;
    
    private DatagramSocket socket;
    
    public LanguageModelClient(String _host, int _port) throws IOException{
        serverAddress = InetAddress.getByName(_host);
        port = _port;
        socket = new DatagramSocket();
    }
    
    public Double query(String sentence) throws IOException{
        byte[] bytesToSend = sentence.getBytes();
        DatagramPacket sendPacket = new DatagramPacket(bytesToSend, bytesToSend.length,
          serverAddress, port);
        DatagramPacket receivePacket = new DatagramPacket(new byte[8], 8);
        
        int tries = 0;
        
        boolean receivedResponse = false;
        
        do {
            socket.send(sendPacket);
            try {
                socket.receive(receivePacket);
                
                if (!receivePacket.getAddress().equals(serverAddress)) {
                }
                else {
                    receivedResponse = true;
                }
            }
            catch (InterruptedIOException e) {
                
                tries +=1;
                System.out.println("Timed out, " + (MAXTRIES - tries) + " more tries...");
            }
        }
        while ((!receivedResponse) && (tries < MAXTRIES));
            
        if (receivedResponse) {
            Double response = ByteBuffer.wrap(receivePacket.getData()).getDouble();
            //System.out.println(response);
            return response;
        }
        else {
            System.out.println("Giving up");
            return null;
        }
       
    }
    
    public static void main(String [] args) throws IOException{
        System.out.println("hello");
        LanguageModelClient client = new LanguageModelClient("localhost", 50001);
        for (int i = 0; i < 1; i++) {
            client.query("Hello world again\n");
        }
    }
}