
import java.util.Hashtable;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;

/**
 * Implements the TranslationModel interface to perform the training
 * and sentence probability evaluation of IBM Model 1.
 */

class Model1 extends TranslationModel {
	
    /**
     * Does precisely bupkis.
     */

    public Model1()
    {
	//
    }

    /**
     * Initializes the given parameters to contain uniform T values
     * for the given corpus. Assumes that the given
     * parameter object is empty to begin with.
     */

    void initialize(ParameterAccessor params, CorpusAccessor corpus)
    {
	//initialize stuff
		
	Hashtable<String, Integer> numAssociatedWords = new Hashtable<String, Integer>();
	HashSet<String> wordPairs = new HashSet<String>();
		
	for (int i=0; i < corpus.getSize(); i++)
	    {
		String[] sourceSentence = corpus.getSentence(i, true);
		String[] destSentence = corpus.getSentence(i, false);
		    
		for (String sourceWord : sourceSentence)
		    {
		        for (String destWord : destSentence) 
			    {
				if (wordPairs.add(destWord + " " + sourceWord)) {
				    if (!numAssociatedWords.containsKey(sourceWord))
					numAssociatedWords.put(sourceWord, new Integer(0));
				    numAssociatedWords.put(sourceWord, new Integer(numAssociatedWords.get(sourceWord).intValue() + 1));
				}
			    }
		    }
	    }
		
	//Enumeration en = numAssociatedWords.keys();
		
	//while(en.hasMoreElements())
	//{
	//String key = (String) en.nextElement();
	//System.out.println("Count for " + key + ": " + numAssociatedWords.get(key).toString());
	//}
		
	for(String wordPair : wordPairs)
	    {
		String[] splitPair = wordPair.split(" ");
		if(splitPair.length != 2)
		    {
		        System.out.println("ERROR!! \"" + wordPair + "\" has wrong number of spaces!");
		        return;
		    }
		int numWords = numAssociatedWords.get(splitPair[1]);
		params.setT(splitPair[0], splitPair[1], 1.0 / numWords);
	    }
		
	/*
	  String[] sourceWords = corpus.getWordList(true);
	  String[] destWords = corpus.getWordList(false);
		
	  double uniformProbability = 1.0 / destWords.length;
		
	  for(String sourceWord : sourceWords)
	  {
	  for(String destWord : destWords)
	  {
	  params.setT(destWord, sourceWord, uniformProbability);
	  }
	  }
	*/
    }

    /**
     * Writes to the parameter object the result of training those
     * parameters for some number of additional iterations based
     * on the contents of the specified corpus. The parameters must
     * have been initialized by the initialize function (and then
     * possibly since trained for some number of iterations).

     @param params the parameters which are trained and updated
     @param corpus the corpus used to guide the training
     @param iters the number of iterations of training to perform
    */
	
    void train(ParameterAccessor params, CorpusAccessor corpus, int iters)
    {
	//REMINDER: should we replace iters with (or additionally offer)
	//a way to cut off training after it has converged to a certain degree?
		
	//get lists of unique words in the corpus
	String[] sourceWords = corpus.getWordList(true);
	String[] destWords = corpus.getWordList(false);
		
	int totalSourceWords = sourceWords.length;
		
	System.out.println("About to start training...");
		
	for(int i = 0; i < iters; i++)
	    {
		int numSourceWord = 0;
		System.out.println("Training Model1 translation model - Iteration " + Integer.toString(i + 1) + " of " + Integer.toString(iters));
		Hashtable<String, Double> tValues = new Hashtable<String, Double>();
		for(String sourceWord : sourceWords)
		    {
			if(numSourceWord % 100 == 0)
			    System.out.println("\tSource word " + Integer.toString(numSourceWord) + " of " + Integer.toString(totalSourceWords));
			//generate all counts for this word
			//corpus, params, dest word list
			Hashtable<String, Double> counts = generateCounts(params, corpus, destWords, sourceWord);
				
			//calculate the normalizing weight for this word
			double normalizationWeight = 0.0;
			for (Double d : counts.values())
			    normalizationWeight += d.doubleValue();
				
			//update the t values for this word and all dest words
			for(String destWord : destWords) {
			    double tempCount = counts.get(destWord).doubleValue();
			    if((tempCount > 0) || (params.getT(destWord, sourceWord) != 0))
				tValues.put(destWord+" "+sourceWord, tempCount / normalizationWeight);
			}
			numSourceWord++;
		    }
		Enumeration en = tValues.keys();
		String destWord;
		String sourceWord;
		while (en.hasMoreElements()) {
                    String pair = (String) en.nextElement();
                    String[] paira = pair.split(" ");
                    destWord = paira[0];
                    sourceWord = paira[1];
                    params.setT(destWord, sourceWord, tValues.get(pair));
		}
	    }
    }

    /**
     * Generates a hash table mapping unique words in the destination
     * language to the "partial counts," the non-normalized updated
     * values for the T probabilities of those destination words
     * given the single, specified source word.

     @param params the current parameters from which the partial
     * counts are generated. The contents of the object are not modified
     * in this function.
     @param corpus the corpus used to find the partial counts
     @param destWords an array of all the unique words in the destination
     * language
     @param sourceWord the single source-language word being associated
     * with all the destination words for the partial counts

     @return the hashtable of partial counts described above
    */
	
    Hashtable<String, Double> generateCounts(ParameterAccessor params, CorpusAccessor corpus, String[] destWords, String sourceWord)
    {
	Hashtable<String, Double> counts = new Hashtable<String, Double>();
	for (String w : destWords) 
	    counts.put(w, new Double(0));
	for (int i=0; i < corpus.getSize(); i++) {
	    String[] sourceSentence = corpus.getSentence(i, true);
	    String[] destSentence = corpus.getSentence(i, false);
	    int sourceWordCount = wordCountInSentence(sourceWord, sourceSentence);
	    if (sourceWordCount == 0) {
		continue;
	    }
	    Hashtable<String, Integer> destWordsAndCounts = wordsAndCountsInSentence(destSentence);
	    Enumeration en = destWordsAndCounts.keys();
	    while (en.hasMoreElements()) {
		String destWord = (String) en.nextElement();
		int destWordCount = destWordsAndCounts.get(destWord);
		double numerator = params.getT(destWord, sourceWord);
		double denominator = 0.0;
		for (String w : sourceSentence) {
		    denominator += params.getT(destWord, w);
		}
				
		//Equation 17 from Brown:
		double partialCount = (numerator / denominator) * destWordCount * sourceWordCount;
		counts.put(destWord, new Double(counts.get(destWord).doubleValue() + partialCount));
	    }
	}
	return counts;
    }

    /**
     * Returns the probability that Model 1 would generate the given
     * destination sentence as a translation of the source sentence.
     * For Model 1, this calculation is performed exactly.

     @param params the parameters used to perform the probability
     * evaluation
     @param destSentence the sentence generated by the source sentence;
     * words must be space-separated
     @param sourceSentence the sentence generating the destination
     * sentence; words must be space-separated

     @return the probability that the given source sentence
     * generates the given destination sentence under Model 1
     */
	
    double sentenceProbability(ParameterAccessor params, String destSentence, String sourceSentence)
    {
	//Note: this probability value does not take sentence lengths into account
	String[] destWords = destSentence.split("\\s+");
	String[] sourceWords = sourceSentence.split("\\s+");
		
	return sentenceProbability(params, destWords, sourceWords);
    }

    /**
     * Returns the probability that Model 1 would generate the given
     * destination sentence as a translation of the source sentence,
     * where the two sentences are represented as arrays of individual
     * words. Being Model 1, this calculation is performed exactly.

     @param params the parameters used to perform the probability
     * evaluation
     @param destSentence the sentence generated by the source sentence,
     * as an array of words
     @param sourceSentence the sentence generating the destination
     * sentence, as an array of words

     @return the probability that the given source sentence
     * generates the given destination sentence under Model 1
     */
	
    double sentenceProbability(ParameterAccessor params, String[] destSentence, String[] sourceSentence)
    {
	//Note: this probability value does not take sentence lengths into account
	
	double product = 1.0;
		
	for(String destWord : destSentence)
	    {
		double sum = 0.0;
		for(String sourceWord : sourceSentence)
		    sum += params.getT(destWord, sourceWord);
		        
		product *= sum;
	    }
		
	return product;
    }	
    //double sentenceAndAlignmentProbability(ParameterAccessor params, String s1, Alignment a, String s2)
    //{
    //
    //}

    /**
     * Returns the number of times that the specified word appears
     * in the given array-of-words sentence.

     @param word the word being looked for
     @param sentence the sentence in which to look for it, as an
     * array of single words

     @return the number of times the word appears in the sentence
    */

    int wordCountInSentence(String word, String[] sentence)
    {
	int count = 0;
	for (String w : sentence) {
	    if (w.equals(word))
		count++;
	}
	return count;
    }

    /**
     * Returns a hash table matching each unique word in the given
     * array-of-words sentence to the number of times it appears
     * in the sentence.

     @param sentence the array-of-words sentence being examined

     @return the hash table described above
    */
	
    Hashtable<String, Integer> wordsAndCountsInSentence(String[] sentence)
    {
	Hashtable<String, Integer> words = new Hashtable<String, Integer>();
	for (String w : sentence) {
	    if (!words.containsKey(w))
		words.put(w, new Integer(1));
	    else
		words.put(w, new Integer(words.get(w).intValue()+1)); // :(
	}
	return words;
    }

    /**
     * Checks whether the T values in the given parameters satisfy the
     * summing-to-one properties of a conditional probability function,
     * given the corpus from which the parameters were trained.
     * Raises an assertion if any of the sums that should equal one are
     * far beyond what floating-point error could explain.

     @param params the parameters being sanity-checked
     @param corpus the corpus used to train the parameters
    */
	
    public static void checkSumToOne(ParameterAccessor params, CorpusAccessor corpus)
    {
	String[] sourceWords = corpus.getWordList(true);
	String[] destWords = corpus.getWordList(false);
		
	for(String sourceWord : sourceWords)
	    {
		double sum = 0.0;
		for(String destWord : destWords)
		    {
			sum += params.getT(destWord, sourceWord);
		    }
		if((sum > 1.01) || (sum < 0.99))
		    {
			System.out.println("Failed on " + sourceWord + " with sum " + Double.toString(sum));
			assert(false);
		    }
	    }
		
    }

    /**
     * Performs unit tests on the Model 1 implementation.
     */
	
    public static void test()
    {
	System.out.println("Testing sum-to-1-iness of initialized params...");
	ParameterAccessor params = new ParametersFromTextFile();
	CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/spanish.txt", "../corpora/textpairs/english.txt");
	Model1 model1 = new Model1();
	model1.initialize(params, corpus);
		
	Model1.checkSumToOne(params, corpus);
	System.out.println("Test successful");
		
	System.out.println("Testing sum-to-1-iness of trained params...");
	model1.train(params, corpus, 10);
	Model1.checkSumToOne(params, corpus);
	System.out.println("Test successful");
		
	System.out.println("Testing word counting...");
	String [] sentence = corpus.getSentence(0, true);
	Hashtable<String, Integer> words = model1.wordsAndCountsInSentence(sentence);
	assert(words.get("the") == 2);
	assert(words.get("English") == 1);
	System.out.println("Test successful");
    }

    /**
     * Tests, trains, or applies Model 1, depending on how we're feeling
     * at the moment. We should probably either make this respond to
     * command-line arguments, or devote this main to testing and
     * add another file for performing Model 1 training.
     */
	
    public static void main(String[] args)
    {
	//test initialize
	Model1.test();
	/*
	  System.out.println("Hello world!");
	  ParametersFromTextFile params = new ParametersFromTextFile();
	  Model1 model1 = new Model1();
		
	  CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.es", "../corpora/textpairs/europarl_2-to-10_stripped.training.es-en.en");
	*/
	/*
	  params.importParams("europarl_2-to-10_stripped.training.es-en.params");
	  //model1.initialize(params, corpus);


	  //System.out.println(params.getT("Spanish", "English"));
	  //System.out.println(params.getT("the", "English"));
	  //params.printT();
	  //String[] testSentence = "This is a test sentence sentence sentence".split("\\s+");
	  //System.out.println("Test sentence: " + Arrays.toString(testSentence));
	  //String testWord = "sentence";
	  //int result = model1.wordCountInSentence(testWord, testSentence);
	  //System.out.println("Result: "+result);
	  //Hashtable<String, Integer> result2 = model1.wordsAndCountsInSentence(testSentence);
	  //System.out.println("Result # 2: "+ result2.toString());
		
	  //Hashtable<String, Double> counts = model1.generateCounts(params, corpus, corpus.getWordList(false), "this");
	  //System.out.println("Counts: " + counts.toString());

		
	  model1.train(params, corpus, 1);
	  //params.printT();

	  System.out.println("Starting export...");
	  params.exportParams("europarl_2-to-10_stripped.training.es-en.params");
	  System.out.println("Done with export...");
		
	  String[] englishWords = corpus.getWordList(true);
	  String[] spanishWords = corpus.getWordList(false);
	  System.out.println("English unique count: " + Integer.toString(englishWords.length));
	  System.out.println("Spanish unique count: " + Integer.toString(spanishWords.length));
		
	*/
		
		
		
	/*
	  System.out.println("Starting import...");
	  params.importParams("../parameters/textfiles/europarl_2-to-10_stripped.training.es-en.model1.5iters");
	  System.out.println("Done with import...");
		
	  String sourceSentence = "today is wednesday";
	  String destSentence = "hoy es miércoles";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	  sourceSentence = "today is thursday";
	  destSentence = "hoy es miércoles";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	  sourceSentence = "today is friday";
	  destSentence = "hoy es miércoles";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	  sourceSentence = "today was wednesday";
	  destSentence = "hoy es miércoles";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	  sourceSentence = "where is the country";
	  destSentence = "donde está el país";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	  sourceSentence = "where is the house";
	  destSentence = "donde está el país";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	  sourceSentence = "where is the country";
	  destSentence = "donde estaba el país";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	  sourceSentence = "where was the country";
	  destSentence = "donde estaba el país";
	  System.out.println("Probability of " + destSentence + " given " + sourceSentence + ": " + Double.toString(model1.sentenceProbability(params, destSentence, sourceSentence)));
	*/
		
	/*
	  for (String e : englishWords) {
	  double sum = 0.0;
	  for (String s : spanishWords) {
	  sum += params.getT(s, e);
	  }
	  System.out.println(e + ": " + Double.toString(sum));
	  }
	*/
		
    }
	
}
