
import java.util.Hashtable;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;

/**
 * Implements the TranslationModel interface to perform the training
 * and sentence probability evaluation of IBM Model 2.
 */

class Triple {

	public int sourceLength;
	public int destLength;
	public int sourcePosition;
	
	public Triple(int inputSourcePosition, int inputSourceLength, int inputDestLength) {
		sourceLength = inputSourceLength;
		destLength = inputDestLength;
		sourcePosition = inputSourcePosition;
	}
}

class Model2 extends TranslationModel {

    public Model2()
    {
    
    }

	/**
	 * Writes to the parameter object the result of training those
	 * parameters for some number of additional iterations based
	 * on the contents of the specified corpus. The parameters must
	 * have been initialized by the initialize function (and then
	 * possibly since trained for some number of iterations).

	 @param params the parameters which are trained and updated
	 @param corpus the corpus used to guide the training
	 @param iters the number of iterations of training to perform
	 */

	/**
	 * Initializes the given parameters to contain uniform T values
	 * for the given corpus. Assumes that the given
	 * parameter object is empty to begin with.
	 */

	void initialize(ParameterAccessor params, CorpusAccessor corpus)
	{
		//initialize stuff
		
		Hashtable<String, Integer> numAssociatedWords = new Hashtable<String, Integer>();
		HashSet<String> wordPairs = new HashSet<String>();
		
		for (int i=0; i < corpus.getSize(); i++)
		{
		    String[] sourceSentence = corpus.getSentence(i, true);
		    String[] destSentence = corpus.getSentence(i, false);
		    
		    for (String sourceWord : sourceSentence)
		    {
		        for (String destWord : destSentence) 
		        {
		            if (wordPairs.add(destWord + " " + sourceWord)) {
		                if (!numAssociatedWords.containsKey(sourceWord))
		                    numAssociatedWords.put(sourceWord, new Integer(0));
		                numAssociatedWords.put(sourceWord, new Integer(numAssociatedWords.get(sourceWord).intValue() + 1));
		            }
		        }
		    }
		}
		
		//Enumeration en = numAssociatedWords.keys();
		
		//while(en.hasMoreElements())
		//{
		    //String key = (String) en.nextElement();
		    //System.out.println("Count for " + key + ": " + numAssociatedWords.get(key).toString());
		//}
		
		for(String wordPair : wordPairs)
		{
		    String[] splitPair = wordPair.split(" ");
		    if(splitPair.length != 2)
		    {
		        System.out.println("ERROR!! \"" + wordPair + "\" has wrong number of spaces!");
		        return;
		    }
		    int numWords = numAssociatedWords.get(splitPair[1]);
		    params.setT(splitPair[0], splitPair[1], 1.0 / numWords);
		}
		
		int counter = 0;
		for (int m=1; m < 15; m++) 
		{
		    for (int l=1; l < 15; l++)
		    {
		        for (int i=0; i < l; i++)
		        {
		            for (int j=0; j < m; j++)
		            {
		                params.setD(i, j, m, l, 1.0/(l+1));
		                counter++;
		            }
		        }
		    }
		}
		System.out.println(counter);
		
		/*
		String[] sourceWords = corpus.getWordList(true);
		String[] destWords = corpus.getWordList(false);
		
		double uniformProbability = 1.0 / destWords.length;
		
		for(String sourceWord : sourceWords)
		{
			for(String destWord : destWords)
			{
				params.setT(destWord, sourceWord, uniformProbability);
			}
		}
		*/
	}

    
	void train(ParameterAccessor params, CorpusAccessor corpus, int iters)
	{
	
		//Model1 model1Instance = new Model1();
	    //model1Instance.initialize(params, corpus);
	    //model1Instance.train(params, corpus, 1);
		//REMINDER: should we replace iters with (or additionally offer)
		//a way to cut off training after it has converged to a certain degree?
		
		//get lists of unique words in the corpus
		String[] sourceWords = corpus.getWordList(true);
		String[] destWords = corpus.getWordList(false);
		
		int totalSourceWords = sourceWords.length;
		
		System.out.println("About to start training...");
		
		for(int i = 0; i < iters; i++)
		{
		    int numSourceWord = 0;
		    System.out.println("Training Model2 translation model - Iteration " + Integer.toString(i + 1) + " of " + Integer.toString(iters));
			for(String sourceWord : sourceWords)
			{
				if(numSourceWord % 100 == 0)
					System.out.println("\tSource word " + Integer.toString(numSourceWord) + " of " + Integer.toString(totalSourceWords));
				//generate all counts for this word
				//corpus, params, dest word list
				Hashtable<String, Double> tCounts = generateTCounts(params, corpus, destWords, sourceWord);
				//System.out.println(tCounts.get("el"));
				
				//calculate the normalizing weight for this word
				double normalizationWeight = 0.0;
				for (Double d : tCounts.values())
					normalizationWeight += d.doubleValue();
				
				//update the t values for this word and all dest words
				//System.out.println(tCounts.size());
				for(String destWord : destWords) {
				    double tempCount = tCounts.get(destWord).doubleValue();
				    //System.out.println(tempCount);
				    if((tempCount > 0) || (params.getT(destWord, sourceWord) != 0))
					    params.setT(destWord, sourceWord, tempCount / normalizationWeight);
				}
				numSourceWord++;
				
			}
			generateDCounts(params, corpus);
			//add training probability between sentence lengths, plus add param accessor getter, setter
		}
	}

	/**
	 * Generates a hash table mapping unique words in the destination
	 * language to the "partial counts," the non-normalized updated
	 * values for the T probabilities of those destination words
	 * given the single, specified source word.

	 @param params the current parameters from which the partial
	 * counts are generated. The contents of the object are not modified
	 * in this function.
	 @param corpus the corpus used to find the partial counts
	 @param destWords an array of all the unique words in the destination
	 * language
	 @param sourceWord the single source-language word being associated
	 * with all the destination words for the partial counts

	 @return the hashtable of partial counts described above
	 */
	
	Hashtable<String, Double> generateTCounts(ParameterAccessor params, CorpusAccessor corpus, String[] destWords, String sourceWord)
	{
		Hashtable<String, Double> counts = new Hashtable<String, Double>();
		for (String w : destWords)
			counts.put(w, new Double(0));
		for (int x=0; x < corpus.getSize(); x++) {
			String[] sourceSentence = corpus.getSentence(x, true);
			String[] destSentence = corpus.getSentence(x, false);
			Hashtable<String, Integer> destWordsAndCounts = wordsAndCountsInSentence(destSentence);
			Enumeration en = destWordsAndCounts.keys();
			while (en.hasMoreElements()) {
				String destWord = (String) en.nextElement();
				double partialCount = 0.0;
				double numeratorTValue = params.getT(destWord, sourceWord);
				for (int j=0; j < sourceSentence.length; j++) {
					
					int cmp = sourceWord.compareTo(sourceSentence[j]);
					if (cmp == 0) {
					
						double subPartialCount = 0.0;
						double denominatorTValue = 0.0;
						/*
						for (int k=0; k < sourceSentence.length; k++) {
							denominatorTValue += params.getT(destWord, sourceSentence[k]);
						}
						*/
						for (int i=0; i < destSentence.length; i++) {
						
							cmp = destWord.compareTo(destSentence[i]);
							if (cmp == 0) {
							
								double numeratorAValue = params.getD(j,i,destSentence.length,sourceSentence.length);
								double denominator = 0.0;
								for (int k=0; k < sourceSentence.length; k++) {
									denominator += params.getT(destWord, sourceSentence[k])*params.getD(k,i,destSentence.length, sourceSentence.length);
								}
								subPartialCount += numeratorTValue*numeratorAValue/denominator;
							}
						}
						partialCount += subPartialCount;
					}
				}
				counts.put(destWord, new Double(counts.get(destWord).doubleValue() + partialCount));
			}
		}
		return counts;
	}
	
	void generateDCounts(ParameterAccessor params, CorpusAccessor corpus)
	{	
		for (int j=0; j < 15; j++) {
				
				Hashtable<Triple, Double> counts = new Hashtable<Triple, Double>();
				/*
				for (int a=1; a < 15; a++) 
				{
					for (int b=1; b < 15; b++)
					{
						for (int c=0; c < b; c++)
						{
							counts.put(new Triple(c, b, a), 0.0);
						}
					}
				}
				*/
			for (int x=0; x < corpus.getSize(); x++) {
				String[] sourceSentence = corpus.getSentence(x, true);
				String[] destSentence = corpus.getSentence(x, false);
				
				if (j < destSentence.length) {
					//double partialCount = 0.0
					for (int i=0; i < sourceSentence.length; i++) {
						Triple trips = new Triple(i,sourceSentence.length, destSentence.length);
						double numerator = params.getT(destSentence[j], sourceSentence[i])*params.getD(i,j,destSentence.length,sourceSentence.length);
						double denominator = 0.0;
						for (int k=0; k < sourceSentence.length; k++) {
							denominator += params.getT(destSentence[j], sourceSentence[k])*params.getD(k,j,destSentence.length,sourceSentence.length);
						}
						if (counts.containsKey(trips)) {
							counts.put(trips, new Double(counts.get(trips).doubleValue() + numerator/denominator));
						} else {
							counts.put(trips, numerator/denominator);
						}
					}
				}
			}
			double normalizationWeight = 0.0;
			for (Double d : counts.values()) {
				normalizationWeight+=d;
			}
			Enumeration en = counts.keys();
			while (en.hasMoreElements()) {
				Triple t = (Triple) en.nextElement();
				double tempCount = counts.get(t).doubleValue();
				if((tempCount > 0) || (params.getD(t.sourcePosition,j,t.destLength,t.sourceLength) != 0))
					params.setD(t.sourcePosition,j,t.destLength,t.sourceLength, tempCount / normalizationWeight);
			}
		}
	}
	
	/**
	 * Returns the number of times that the specified word appears
	 * in the given array-of-words sentence.

	 @param word the word being looked for
	 @param sentence the sentence in which to look for it, as an
	 * array of single words

	 @return the number of times the word appears in the sentence
	 */

	int wordCountInSentence(String word, String[] sentence)
	{
		int count = 0;
		for (String w : sentence) {
			if (w.equals(word))
				count++;
		}
		return count;
	}

	/**
	 * Returns a hash table matching each unique word in the given
	 * array-of-words sentence to the number of times it appears
	 * in the sentence.

	 @param sentence the array-of-words sentence being examined

	 @return the hash table described above
	 */
	
	Hashtable<String, Integer> wordsAndCountsInSentence(String[] sentence)
	{
		Hashtable<String, Integer> words = new Hashtable<String, Integer>();
		for (String w : sentence) {
			if (!words.containsKey(w))
				words.put(w, new Integer(1));
			else
				words.put(w, new Integer(words.get(w).intValue()+1)); // :(
		}
		return words;
	}
	

	/**
	 * Returns the probability that Model 1 would generate the given
	 * destination sentence as a translation of the source sentence.
	 * For Model 1, this calculation is performed exactly.

	 @param params the parameters used to perform the probability
	 * evaluation
	 @param destSentence the sentence generated by the source sentence;
	 * words must be space-separated
	 @param sourceSentence the sentence generating the destination
	 * sentence; words must be space-separated

	 @return the probability that the given source sentence
	 * generates the given destination sentence under Model 1
	 */
	
	double sentenceProbability(ParameterAccessor params, String destSentence, String sourceSentence)
	{
		//Note: this probability value does not take sentence lengths into account
		String[] destWords = destSentence.split("\\s+");
		String[] sourceWords = sourceSentence.split("\\s+");
		
		return sentenceProbability(params, destWords, sourceWords);
	}

	/**
	 * Returns the probability that Model 1 would generate the given
	 * destination sentence as a translation of the source sentence,
	 * where the two sentences are represented as arrays of individual
	 * words. Being Model 1, this calculation is performed exactly.

	 @param params the parameters used to perform the probability
	 * evaluation
	 @param destSentence the sentence generated by the source sentence,
	 * as an array of words
	 @param sourceSentence the sentence generating the destination
	 * sentence, as an array of words

	 @return the probability that the given source sentence
	 * generates the given destination sentence under Model 1
	 */
	
	double sentenceProbability(ParameterAccessor params, String[] destSentence, String[] sourceSentence)
	{
		//Note: this probability value does not take sentence lengths into account
	
		double product = 1.0;
		
		for(String destWord : destSentence)
		{
		    double sum = 0.0;
		    for(String sourceWord : sourceSentence)
		        sum += params.getT(destWord, sourceWord);
		        
		    product *= sum;
		}
		
		return product;
	}
	
	public static void main(String[] args)
	{
		//test initialize
		
		System.out.println("Hello world!");
		ParametersFromTextFile params2 = new ParametersFromTextFile();
		ParametersFromTextFile params1 = new ParametersFromTextFile();
		Model1 model1 = new Model1();
		Model2 model2 = new Model2();
		
		CorpusFromTextPair corpus = new CorpusFromTextPair("../corpora/textpairs/thousand.es", "../corpora/textpairs/thousand.en");

		model2.initialize(params2, corpus);
		//model1.initialize(params1, corpus);
		System.out.println("initialized");
        model2.train(params2, corpus, 5);
        //model1.train(params1, corpus, 5);
		//System.out.println(params2.getT("the", "el"));
		//System.out.println(params1.getT("the", "el"));
		params2.printT();
		

		
	}
	
}